from __future__ import annotations

from json import dumps

from sympy import latex as to_latex


def format_result(value, format_mode: str) -> str:
    if format_mode == "plain":
        return str(value)
    if format_mode == "pretty":
        from sympy import pretty as to_pretty

        return to_pretty(value)
    rendered = to_latex(value)
    if format_mode == "latex-inline":
        return f"${rendered}$"
    if format_mode == "latex-block":
        return f"$$\n{rendered}\n$$"
    return rendered


def format_json_result(
    expr: str,
    relaxed: bool,
    value,
    *,
    normalize_expression_fn,
) -> str:
    normalized = normalize_expression_fn(expr, relaxed=relaxed)
    payload = {"input": expr, "parsed": normalized, "result": str(value)}
    return dumps(payload, separators=(",", ":"))


def render_value(
    value,
    *,
    format_mode: str,
    expr: str,
    relaxed: bool,
    normalize_expression_fn,
    parsed_expr: str | None = None,
) -> str:
    if format_mode == "json":
        if parsed_expr is None:
            return format_json_result(
                expr,
                relaxed,
                value,
                normalize_expression_fn=normalize_expression_fn,
            )
        payload = {"input": expr, "parsed": parsed_expr, "result": str(value)}
        return dumps(payload, separators=(",", ":"))
    return format_result(value, format_mode)



import pytest
from sympy import Eq, Function, Symbol

from calc.core import evaluate
from calc.ode import evaluate_ode_alias, infer_ode_dependent, split_top_level_commas


def test_split_top_level_commas_handles_nested_parens():
    assert split_top_level_commas("y' = y, y(0)=1") == ["y' = y", "y(0)=1"]
    assert split_top_level_commas("Eq(d(y(x), x), y(x)), Eq(y(0), 1)") == [
        "Eq(d(y(x), x), y(x))",
        "Eq(y(0), 1)",
    ]


def test_infer_ode_dependent_returns_none_without_applied_function():
    x = Symbol("x")
    assert infer_ode_dependent(Eq(x, 1)) is None


def test_evaluate_ode_alias_success_without_ics():
    value, parsed = evaluate_ode_alias(
        "ode y' = y",
        evaluate_fn=evaluate,
        relaxed=True,
        simplify_output=True,
        session_locals={},
    )
    assert str(value) == "Eq(y(x), C1*exp(x))"
    assert parsed.startswith("dsolve(Eq(")


def test_evaluate_ode_alias_success_with_ics():
    value, parsed = evaluate_ode_alias(
        "ode y' = y, y(0)=1",
        evaluate_fn=evaluate,
        relaxed=True,
        simplify_output=True,
        session_locals={},
    )
    assert str(value) == "Eq(y(x), exp(x))"
    assert "ics={y(0): 1}" in parsed


def test_evaluate_ode_alias_accepts_prime_ic_and_implicit_dependent_term():
    value, parsed = evaluate_ode_alias(
        "ode y'' + 9*dy/dx + 20y = 0, y(0)=1, y'(0)=0",
        evaluate_fn=evaluate,
        relaxed=True,
        simplify_output=True,
        session_locals={},
    )
    assert str(value) == "Eq(y(x), (5 - 4*exp(-x))*exp(-4*x))"
    assert "ics={y(0): 1, Subs(Derivative(y(x), x), x, 0): 0}" in parsed


def test_evaluate_ode_alias_raises_on_empty_body():
    with pytest.raises(ValueError, match="ode expects an equation"):
        evaluate_ode_alias(
            "ode ",
            evaluate_fn=evaluate,
            relaxed=True,
            simplify_output=True,
            session_locals={},
        )


def test_evaluate_ode_alias_raises_on_empty_pieces():
    with pytest.raises(ValueError, match="ode expects an equation"):
        evaluate_ode_alias(
            "ode , ,",
            evaluate_fn=evaluate,
            relaxed=True,
            simplify_output=True,
            session_locals={},
        )


def test_evaluate_ode_alias_requires_equation_and_dependent_function():
    def fake_eval(expr: str, **kwargs):
        if expr == "x+1":
            return 3
        return Eq(Symbol("x"), 1)

    with pytest.raises(ValueError, match="ode expects an equation"):
        evaluate_ode_alias(
            "ode x+1",
            evaluate_fn=fake_eval,
            relaxed=True,
            simplify_output=True,
            session_locals={},
        )

    with pytest.raises(ValueError, match="could not infer dependent function"):
        evaluate_ode_alias(
            "ode Eq(x, 1)",
            evaluate_fn=fake_eval,
            relaxed=True,
            simplify_output=True,
            session_locals={},
        )


def test_evaluate_ode_alias_validates_initial_conditions_and_passes_flags():
    y = Function("y")
    x = Symbol("x")
    calls: list[tuple[str, bool, bool, dict]] = []

    def fake_eval(expr: str, **kwargs):
        calls.append((expr, kwargs["relaxed"], kwargs["simplify_output"], kwargs["session_locals"]))
        if expr == "y' = y":
            return Eq(y(x), y(x).diff(x))
        return 1

    with pytest.raises(ValueError, match="initial condition must be an equation"):
        evaluate_ode_alias(
            "ode y' = y, 1",
            evaluate_fn=fake_eval,
            relaxed=False,
            simplify_output=False,
            session_locals={"a": 1},
        )

    assert calls[0] == ("y' = y", False, False, {"a": 1})
    assert calls[1] == ("1", False, False, {"a": 1})


def test_evaluate_ode_alias_rejects_boolean_initial_condition():
    with pytest.raises(ValueError, match="initial condition reduced to a boolean"):
        evaluate_ode_alias(
            "ode y' = y, 0=0",
            evaluate_fn=evaluate,
            relaxed=True,
            simplify_output=True,
            session_locals={},
        )



from __future__ import annotations

import os
import re
import shutil
import subprocess
import sys
from importlib import import_module
from importlib.metadata import PackageNotFoundError, version as package_version
from urllib.parse import quote_plus
from urllib.request import urlopen

from sympy import Eq
from sympy.matrices.matrixbase import MatrixBase

from .core import evaluate, normalize_expression
from .diagnostics import (
    hint_for_error as _hint_for_error,
    parse_explanation,
    relaxed_rewrite_messages,
    should_print_wolfram_hint,
)
from .options import COLOR_MODES, CLIOptions, parse_options
from .ode import (
    evaluate_ode_alias as evaluate_ode_alias_impl,
    infer_ode_dependent as infer_ode_dependent_impl,
    split_top_level_commas as split_top_level_commas_impl,
)
from .repl import (
    handle_repl_command as handle_repl_command_impl,
    try_parse_repl_inline_options as try_parse_repl_inline_options_impl,
    tutorial_command as tutorial_command_impl,
)
from .render import (
    format_json_result as format_json_result_impl,
    format_result as format_result_impl,
    render_value as render_value_impl,
)
from .updates import (
    compare_versions as _compare_versions_impl,
    latest_pypi_version as _latest_pypi_version_impl,
    repl_startup_update_status_lines,
    update_status_lines,
)

PACKAGE_NAME = "philcalc"
CLI_NAME = "phil"
UPDATE_CMD = "uv tool upgrade philcalc"


def _calc_version() -> str:
    try:
        return package_version(PACKAGE_NAME)
    except PackageNotFoundError:
        return "dev"


VERSION = _calc_version()
PROMPT = f"{CLI_NAME}> "
HELP_TEXT = (
    f"{CLI_NAME} v{VERSION} - symbolic CLI calculator\n"
    "\n"
    "usage:\n"
    f"  {CLI_NAME} [--format MODE] [--latex|--latex-inline|--latex-block] [--strict] [--no-simplify] [--explain-parse] [--wa] [--copy-wa] [--color MODE] '<expression>'\n"
    f"  {CLI_NAME}\n"
    f"  {CLI_NAME} :examples\n"
    "\n"
    "options:\n"
    "  --format MODE   output mode: plain, pretty, latex, latex-inline, latex-block, json\n"
    "  --latex         print raw LaTeX (no delimiters)\n"
    "  --latex-inline  print LaTeX wrapped as $...$\n"
    "  --latex-block   print LaTeX wrapped as $$...$$\n"
    "  --strict        disable relaxed input parsing\n"
    "  --no-simplify   skip simplify() on parsed expressions\n"
    "  --explain-parse show normalized expression on stderr\n"
    "  --wa            always print WolframAlpha equivalent link\n"
    "  --copy-wa       copy WolframAlpha link to clipboard when shown\n"
    "  --color MODE    diagnostics color: auto, always, never\n"
    "\n"
    "upgrade:\n"
    f"  {UPDATE_CMD}\n"
    "\n"
    "repl commands:\n"
    "  :h, :help      show this help\n"
    "  ?, ??, ???     progressive help chain (discover more features)\n"
    "  :examples      show example expressions\n"
    "  :tutorial      show guided tour for new users\n"
    "  :ode           show ODE quick reference and templates\n"
    "  :linalg, :la   show linear algebra quick reference and templates\n"
    "  :next          next tutorial step (after :tutorial)\n"
    "  :repeat        repeat current tutorial step\n"
    "  :done          exit tutorial mode\n"
    "  :v, :version   show version\n"
    "  :update, :check  check current vs latest version\n"
    "  :q, :quit, :x  quit\n"
    "\n"
    "quick examples:\n"
    "  (1 - 25e^5)e^{-5t} + (25e^5 - 1)t e^{-5t} + t e^{-5t} ln(t)\n"
    "  10^100000 + 1 - 10^100000\n"
    "  sin^2(x) + cos^2(x)\n"
    "  msolve(Matrix([[2,1],[1,3]]), Matrix([1,2]))\n"
    "  d(x^3 + 2*x, x)\n"
    "  int(sin(x), x)\n"
    "  solve(x^2 - 4, x)\n"
    "  N(pi, 20)"
)
HELP_CHAIN_TEXT = (
    "help chain:\n"
    "  ?    = standard help\n"
    "  ??   = power-user shortcuts\n"
    "  ???  = capability demos\n"
)
HELP_POWER_TEXT = (
    "power-user shortcuts:\n"
    "  use inline REPL options: --latex d(x^2, x)\n"
    "  inspect parser rewrites: --explain-parse 'sinx'\n"
    "  machine-readable output: --format json 'sinx'\n"
    "  fast ODE solve alias: ode y' = y, y(0)=1\n"
    "  coefficient symbols inline: S('A'), S('B'), S('C')\n"
    "  session memory: ans + 1, A = Matrix([[1,2],[3,4]])\n"
    "hint: try ??? for capability demos"
)
HELP_DEMO_TEXT = (
    "capability demos:\n"
    "  exact huge integer arithmetic:\n"
    "    10^100000 + 1 - 10^100000\n"
    "    -> 1\n"
    "  exact rational arithmetic:\n"
    "    1/3 + 1/6\n"
    "    -> 1/2\n"
    "  symbolic workflow:\n"
    "    d(x^3 + 2*x, x), int(sin(x), x), solve(x^2 - 4, x)\n"
)
EXAMPLES_TEXT = (
    "examples:\n"
    "  1/3 + 1/6\n"
    "  10^100000 + 1 - 10^100000\n"
    "  sin^2(x) + cos^2(x)\n"
    "  msolve(Matrix([[2,1],[1,3]]), Matrix([1,2]))\n"
    "  d(x^3 + 2*x, x)\n"
    "  int(sin(x), x)\n"
    "  solve(x^2 - 4, x)\n"
    "  solve(2*x^2+43*x+22-(S('A')*(x-7)^2+S('B')*(x+8)*(x-7)+S('C')*(x+8)), (S('A'), S('B'), S('C')))\n"
    "  (854/2197)e^{8t}+(1343/2197)e^{-5t}+((9/26)t^2 -(9/169)t)e^{8t}\n"
    "  dsolve(Eq(f(t).diff(t), f(t)), f(t))\n"
    "  N(pi, 20)"
)
TUTORIAL_TEXT = (
    "guided tour:\n"
    "  1) Start: phil '2+2'\n"
    "  2) Core ops: phil 'd(x^3 + 2*x, x)' / phil 'int(sin(x), x)' / phil 'solve(x^2 - 4, x)'\n"
    "  3) REPL: phil, then try d(x^2, x), A=Matrix([[1,2],[3,4]]), det(A), ans+1\n"
    "  4) ODE input: dy/dx = y ; y' = y ; \\frac{dy}{dx} = y\n"
    "  5) Solve ODE: dsolve(Eq(d(y(x), x), y(x)), y(x))\n"
    "  6) LaTeX style: $d(x^2, x)$ ; \\sin(x)^2 + \\cos(x)^2\n"
    "  7) Exactness demo: phil '10^100000 + 1 - 10^100000'\n"
    "  8) Use :examples and ? / ?? / ??? for more patterns\n"
    "full tutorial: TUTORIAL.md"
)
ODE_TEXT = (
    "ode quick reference:\n"
    "  quick start (human style):\n"
    "    ode y' = y\n"
    "    ode y'' + y = 0\n"
    "    ode y' = y, y(0)=1\n"
    "    ode y'' + 9*y' + 20*y = 0, y(0)=1, y'(0)=0\n"
    "\n"
    "what you can type:\n"
    "  dy/dx = y\n"
    "  y' = y\n"
    "  y'(0)=0\n"
    "  \\frac{dy}{dx} = y\n"
    "  d(y(x), x).subs(x, 0) = 0\n"
    "\n"
    "internal equivalent (advanced):\n"
    "  dsolve(Eq(d(y(x), x), y(x)), y(x))\n"
    "\n"
    "notes:\n"
    "  Eq(...) is equation form (not assignment)\n"
    "  write multiplication explicitly in ODEs (use 20*y, not 20y)\n"
    "  y(x) is dependent-function notation used internally by dsolve\n"
)
LINALG_TEXT = (
    "linear algebra quick reference:\n"
    "  quick start:\n"
    "    linalg solve A=[[2,1],[1,3]] b=[1,2]\n"
    "    linalg rref A=[[1,2],[2,4]]\n"
    "    msolve(Matrix([[2,1],[1,3]]), Matrix([1,2]))\n"
    "    rref(Matrix([[1,2],[2,4]]))\n"
    "    nullspace(Matrix([[1,2],[2,4]]))\n"
    "\n"
    "symbolic systems:\n"
    "  linsolve((Eq(2*x + y, 1), Eq(x + 3*y, 2)), (x, y))\n"
    "\n"
    "notes:\n"
    "  use Matrix([...]) for vectors and matrices\n"
    "  msolve(A, b) expects square, invertible A for LUsolve\n"
)
TUTORIAL_STEPS = (
    "step 1/6\n  run: 1/3 + 1/6\n  expect: 1/2",
    "step 2/6\n  run: d(x^3 + 2*x, x)\n  expect: 3*x**2 + 2",
    "step 3/6\n  run: int(sin(x), x)\n  expect: -cos(x)",
    "step 4/6\n  run: dy/dx = y\n  expect: Eq(y(x), Derivative(y(x), x))",
    "step 5/6\n  run: dsolve(Eq(d(y(x), x), y(x)), y(x))\n  expect: Eq(y(x), C1*exp(x))",
    "step 6/6\n  run: --latex d(x^2, x)\n  expect: 2 x",
)
COLOR_MODES = {"auto", "always", "never"}
ANSI_COLORS = {
    "red": "\033[31m",
    "yellow": "\033[33m",
    "bold": "\033[1m",
}
ANSI_RESET = "\033[0m"


def _wolframalpha_url(expr: str) -> str:
    return f"https://www.wolframalpha.com/input?i={quote_plus(expr)}"


def _is_complex_expression(expr: str) -> bool:
    if len(expr) >= 40:
        return True
    markers = ("d(", "int(", "solve(", "dsolve(", "Eq(", "ln(", "log(", "^", "{", "}", "e^{")
    return any(marker in expr for marker in markers)


def _should_use_color(stream, color_mode: str) -> bool:
    if color_mode not in COLOR_MODES:
        return False
    if color_mode == "never":
        return False
    if color_mode == "always":
        return True
    if os.getenv("NO_COLOR") is not None:
        return False
    if os.getenv("TERM") == "dumb":
        return False
    return stream.isatty()


def _style(text: str, *, color: str, stream, color_mode: str) -> str:
    if not _should_use_color(stream, color_mode):
        return text
    code = ANSI_COLORS.get(color)
    if code is None:
        return text
    return f"{code}{text}{ANSI_RESET}"


def _copy_to_clipboard(text: str) -> bool:
    commands = (
        ["pbcopy"],
        ["wl-copy"],
        ["xclip", "-selection", "clipboard"],
        ["xsel", "--clipboard", "--input"],
        ["clip"],
    )
    for cmd in commands:
        if shutil.which(cmd[0]) is None:
            continue
        try:
            subprocess.run(cmd, input=text, text=True, check=True, capture_output=True)
            return True
        except Exception:
            continue
    return False


def _print_wolfram_hint(expr: str, copy_link: bool = False, color_mode: str = "auto") -> None:
    url = _wolframalpha_url(expr)
    # Always print raw URL for maximum terminal compatibility (e.g., iTerm2 auto-linking).
    print(
        _style(f"hint: try WolframAlpha: {url}", color="yellow", stream=sys.stderr, color_mode=color_mode),
        file=sys.stderr,
    )
    if copy_link:
        if _copy_to_clipboard(url):
            print(
                _style("hint: WolframAlpha link copied to clipboard", color="yellow", stream=sys.stderr, color_mode=color_mode),
                file=sys.stderr,
            )
        else:
            print(
                _style("hint: clipboard copy unavailable on this system", color="yellow", stream=sys.stderr, color_mode=color_mode),
                file=sys.stderr,
            )


def _print_error(
    exc: Exception,
    expr: str | None = None,
    color_mode: str = "auto",
    session_locals: dict | None = None,
) -> None:
    print(_style(f"E: {exc}", color="red", stream=sys.stderr, color_mode=color_mode), file=sys.stderr)
    hint = _hint_for_error(str(exc), expr=expr, session_locals=session_locals)
    if hint:
        print(_style(f"hint: {hint}", color="yellow", stream=sys.stderr, color_mode=color_mode), file=sys.stderr)
    if expr and should_print_wolfram_hint(exc):
        _print_wolfram_hint(expr, color_mode=color_mode)


def _print_relaxed_rewrite_hints(expr: str, relaxed: bool, color_mode: str) -> None:
    for message in relaxed_rewrite_messages(expr, relaxed):
        print(
            _style(
                f"hint: {message}",
                color="yellow",
                stream=sys.stderr,
                color_mode=color_mode,
            ),
            file=sys.stderr,
        )


def _print_parse_explanation(expr: str, relaxed: bool, enabled: bool, color_mode: str) -> None:
    message = parse_explanation(expr, relaxed, enabled)
    if not message:
        return
    print(_style(f"hint: {message}", color="yellow", stream=sys.stderr, color_mode=color_mode), file=sys.stderr)


def _format_result(value, format_mode: str) -> str:
    return format_result_impl(value, format_mode)


def _format_json_result(expr: str, relaxed: bool, value) -> str:
    return format_json_result_impl(
        expr,
        relaxed,
        value,
        normalize_expression_fn=normalize_expression,
    )


def _render_value(value, *, format_mode: str, expr: str, relaxed: bool, parsed_expr: str | None = None) -> str:
    return render_value_impl(
        value,
        format_mode=format_mode,
        expr=expr,
        relaxed=relaxed,
        normalize_expression_fn=normalize_expression,
        parsed_expr=parsed_expr,
    )


def _split_top_level_commas(text: str) -> list[str]:
    return split_top_level_commas_impl(text)


def _infer_ode_dependent(eq_value: Eq):
    return infer_ode_dependent_impl(eq_value)


def _evaluate_ode_alias(
    expr: str,
    *,
    relaxed: bool,
    simplify_output: bool,
    session_locals: dict | None = None,
):
    return evaluate_ode_alias_impl(
        expr,
        evaluate_fn=evaluate,
        relaxed=relaxed,
        simplify_output=simplify_output,
        session_locals=session_locals,
    )


def _consume_bracket_literal(text: str, start: int) -> tuple[str, int]:
    if start >= len(text) or text[start] != "[":
        raise ValueError("expected bracketed literal like [[...]]")
    depth = 0
    idx = start
    while idx < len(text):
        ch = text[idx]
        if ch == "[":
            depth += 1
        elif ch == "]":
            depth -= 1
            if depth == 0:
                return text[start : idx + 1], idx + 1
        idx += 1
    raise ValueError("unclosed bracket literal; expected closing ']'")


def _parse_linalg_keyed_literals(text: str, required_keys: set[str]) -> dict[str, str]:
    idx = 0
    parsed: dict[str, str] = {}
    while idx < len(text):
        while idx < len(text) and text[idx] in {",", " "}:
            idx += 1
        while idx < len(text) and text[idx].isspace():
            idx += 1
        if idx >= len(text):
            break
        key_start = idx
        while idx < len(text) and text[idx].isalpha():
            idx += 1
        key = text[key_start:idx]
        if key not in required_keys:
            expected = ", ".join(sorted(required_keys))
            raise ValueError(f"unknown linalg parameter '{key}'; expected: {expected}")
        if key in parsed:
            raise ValueError(f"duplicate linalg parameter '{key}'")
        while idx < len(text) and text[idx].isspace():
            idx += 1
        if idx >= len(text) or text[idx] != "=":
            raise ValueError(f"linalg parameter '{key}' must use '='")
        idx += 1
        while idx < len(text) and text[idx].isspace():
            idx += 1
        literal, idx = _consume_bracket_literal(text, idx)
        parsed[key] = literal

    missing = sorted(required_keys - set(parsed))
    if missing:
        missing_text = ", ".join(missing)
        raise ValueError(f"missing linalg parameter(s): {missing_text}")
    return parsed


def _evaluate_linalg_alias(
    expr: str,
    *,
    relaxed: bool,
    simplify_output: bool,
    session_locals: dict | None = None,
):
    body = expr[7:].strip()
    if not body:
        raise ValueError("linalg expects a subcommand: solve or rref")
    pieces = body.split(maxsplit=1)
    subcommand = pieces[0].lower()
    rest = pieces[1] if len(pieces) > 1 else ""

    if subcommand == "solve":
        params = _parse_linalg_keyed_literals(rest, {"A", "b"})
        matrix_text = params["A"]
        rhs_text = params["b"]
        matrix_value = evaluate(
            f"Matrix({matrix_text})",
            relaxed=relaxed,
            session_locals=session_locals,
            simplify_output=simplify_output,
        )
        rhs_value = evaluate(
            f"Matrix({rhs_text})",
            relaxed=relaxed,
            session_locals=session_locals,
            simplify_output=simplify_output,
        )
        if not isinstance(matrix_value, MatrixBase) or not isinstance(rhs_value, MatrixBase):
            raise ValueError("linalg solve expects matrix literals for A and b")
        if matrix_value.rows != matrix_value.cols:
            raise ValueError("linalg solve expects square A")
        if rhs_value.cols != 1:
            raise ValueError("linalg solve expects b as a column vector, e.g. b=[1,2]")
        if rhs_value.rows != matrix_value.rows:
            raise ValueError("linalg solve expects len(b) to match rows of A")
        result = matrix_value.LUsolve(rhs_value)
        parsed_expr = f"msolve(Matrix({matrix_text}), Matrix({rhs_text}))"
        return result, parsed_expr

    if subcommand == "rref":
        params = _parse_linalg_keyed_literals(rest, {"A"})
        matrix_text = params["A"]
        matrix_value = evaluate(
            f"Matrix({matrix_text})",
            relaxed=relaxed,
            session_locals=session_locals,
            simplify_output=simplify_output,
        )
        if not isinstance(matrix_value, MatrixBase):
            raise ValueError("linalg rref expects a matrix literal for A")
        result = matrix_value.rref()
        parsed_expr = f"rref(Matrix({matrix_text}))"
        return result, parsed_expr

    raise ValueError("unknown linalg subcommand; use 'solve' or 'rref'")


def _latest_pypi_version() -> str | None:
    return _latest_pypi_version_impl(PACKAGE_NAME, urlopen_fn=urlopen)


def _compare_versions(current: str, latest: str) -> int | None:
    return _compare_versions_impl(current, latest)


def _print_update_status() -> None:
    latest = None if VERSION == "dev" else _latest_pypi_version()
    lines = update_status_lines(
        VERSION,
        latest,
        UPDATE_CMD,
        compare_fn=_compare_versions,
    )
    for line in lines:
        print(line)


def _print_repl_startup_update_status() -> None:
    # Only auto-check when actually interactive to avoid noisy/non-deterministic
    # behavior in piped/scripted REPL sessions.
    if not sys.stdin.isatty():
        return
    latest = None if VERSION == "dev" else _latest_pypi_version()
    lines = repl_startup_update_status_lines(
        VERSION,
        latest,
        UPDATE_CMD,
        compare_fn=_compare_versions,
    )
    for line in lines:
        print(line)


def _parse_options(args: list[str]) -> CLIOptions:
    return parse_options(args, help_text=HELP_TEXT)


def _handle_repl_command(expr: str, color_mode: str = "auto") -> bool:
    return handle_repl_command_impl(
        expr,
        help_text=HELP_TEXT,
        help_chain_text=HELP_CHAIN_TEXT,
        help_power_text=HELP_POWER_TEXT,
        help_demo_text=HELP_DEMO_TEXT,
        examples_text=EXAMPLES_TEXT,
        tutorial_text=TUTORIAL_TEXT,
        ode_text=ODE_TEXT,
        linalg_text=LINALG_TEXT,
        cli_name=CLI_NAME,
        version=VERSION,
        print_update_status=_print_update_status,
        style_fn=_style,
        color_mode=color_mode,
        stderr=sys.stderr,
    )


def _print_tutorial_step(index: int) -> None:
    print(TUTORIAL_STEPS[index])


def _tutorial_command(expr: str, state: dict | None) -> bool:
    return tutorial_command_impl(
        expr,
        state,
        tutorial_steps=TUTORIAL_STEPS,
        print_tutorial_step=_print_tutorial_step,
        stderr=sys.stderr,
    )


def _try_parse_repl_inline_options(expr: str):
    return try_parse_repl_inline_options_impl(
        expr,
        cli_name=CLI_NAME,
        parse_options_fn=_parse_options,
    )


def _configure_repl_line_editing() -> bool:
    if not sys.stdin.isatty():
        return False
    try:
        readline = import_module("readline")
    except Exception:
        return False
    try:
        doc = getattr(readline, "__doc__", "") or ""
        if "libedit" in doc:
            readline.parse_and_bind("bind ^I rl_complete")
        else:
            readline.parse_and_bind("tab: complete")
    except Exception:
        # Arrow key history/editing still works without completion binding.
        pass
    return True


def _execute_expression(
    expr: str,
    *,
    format_mode: str,
    relaxed: bool,
    simplify_output: bool,
    explain_parse: bool,
    always_wa: bool,
    copy_wa: bool,
    color_mode: str,
    session_locals: dict | None = None,
) -> None:
    is_ode_alias = expr.strip().lower().startswith("ode ")
    is_linalg_alias = expr.strip().lower().startswith("linalg ")
    if is_ode_alias:
        value, parsed_expr = _evaluate_ode_alias(
            expr,
            relaxed=relaxed,
            simplify_output=simplify_output,
            session_locals=session_locals,
        )
        if explain_parse:
            print(_style(f"hint: parsed as: {parsed_expr}", color="yellow", stream=sys.stderr, color_mode=color_mode), file=sys.stderr)
        if format_mode == "plain" and isinstance(value, Eq):
            rendered = f"{value.lhs} = {value.rhs}"
        else:
            rendered = _render_value(
                value,
                format_mode=format_mode,
                expr=expr,
                relaxed=relaxed,
                parsed_expr=parsed_expr,
            )
    elif is_linalg_alias:
        value, parsed_expr = _evaluate_linalg_alias(
            expr,
            relaxed=relaxed,
            simplify_output=simplify_output,
            session_locals=session_locals,
        )
        if explain_parse:
            print(_style(f"hint: parsed as: {parsed_expr}", color="yellow", stream=sys.stderr, color_mode=color_mode), file=sys.stderr)
        rendered = _render_value(
            value,
            format_mode=format_mode,
            expr=expr,
            relaxed=relaxed,
            parsed_expr=parsed_expr,
        )
    else:
        _print_relaxed_rewrite_hints(expr, relaxed, color_mode)
        _print_parse_explanation(expr, relaxed, explain_parse, color_mode)
        value = evaluate(
            expr,
            relaxed=relaxed,
            session_locals=session_locals,
            simplify_output=simplify_output,
        )
        rendered = _render_value(value, format_mode=format_mode, expr=expr, relaxed=relaxed)
    print(rendered)
    if always_wa or _is_complex_expression(expr):
        _print_wolfram_hint(expr, copy_link=copy_wa, color_mode=color_mode)


def run(argv: list[str] | None = None) -> int:
    args = sys.argv[1:] if argv is None else argv

    try:
        options = _parse_options(args)
    except SystemExit:
        return 0
    except Exception as exc:
        _print_error(exc, color_mode="auto")
        return 1

    format_mode = options.format_mode
    relaxed = options.relaxed
    simplify_output = options.simplify_output
    explain_parse = options.explain_parse
    always_wa = options.always_wa
    copy_wa = options.copy_wa
    color_mode = options.color_mode
    remaining = list(options.remaining)

    if remaining:
        expr = " ".join(remaining)
        if expr == "?":
            print(HELP_TEXT)
            print(HELP_CHAIN_TEXT)
            return 0
        if expr == "??":
            print(HELP_POWER_TEXT)
            return 0
        if expr == "???":
            print(HELP_DEMO_TEXT)
            return 0
        if expr == ":examples":
            print(EXAMPLES_TEXT)
            return 0
        if expr == ":ode":
            print(ODE_TEXT)
            return 0
        if expr in {":linalg", ":la"}:
            print(LINALG_TEXT)
            return 0
        if expr in {":tutorial", ":tour"}:
            print(TUTORIAL_TEXT)
            return 0
        if expr in {":v", ":version"}:
            print(f"{CLI_NAME} v{VERSION}")
            return 0
        if expr in {":update", ":check"}:
            _print_update_status()
            return 0
        try:
            _execute_expression(
                expr,
                format_mode=format_mode,
                relaxed=relaxed,
                simplify_output=simplify_output,
                explain_parse=explain_parse,
                always_wa=always_wa,
                copy_wa=copy_wa,
                color_mode=color_mode,
            )
            return 0
        except Exception as exc:
            _print_error(exc, expr, color_mode=color_mode)
            return 1

    print(f"{CLI_NAME} v{VERSION} REPL. :h help, :q quit, Ctrl-D exit.")
    if not _configure_repl_line_editing() and sys.stdin.isatty():
        print(
            "hint: line editing unavailable (arrow keys/history may print escape codes); "
            "install Python readline support",
            file=sys.stderr,
        )
    _print_repl_startup_update_status()
    session_locals: dict = {}
    repl_format_mode = format_mode
    repl_relaxed = relaxed
    repl_simplify_output = simplify_output
    repl_explain_parse = explain_parse
    repl_always_wa = always_wa
    repl_copy_wa = copy_wa
    repl_color_mode = color_mode
    tutorial_state = {"active": False, "index": 0}
    expr: str | None = None
    while True:
        try:
            expr = input(PROMPT).strip()
            if not expr:
                continue
            if _tutorial_command(expr, tutorial_state):
                continue
            if _handle_repl_command(expr, color_mode=repl_color_mode):
                continue
            parsed_inline = _try_parse_repl_inline_options(expr)
            if parsed_inline is not None:
                repl_format_mode = parsed_inline.format_mode
                repl_relaxed = parsed_inline.relaxed
                repl_simplify_output = parsed_inline.simplify_output
                repl_explain_parse = parsed_inline.explain_parse
                repl_always_wa = parsed_inline.always_wa
                repl_copy_wa = parsed_inline.copy_wa
                repl_color_mode = parsed_inline.color_mode
                if not parsed_inline.remaining:
                    print("hint: REPL options updated for this session", file=sys.stderr)
                    continue
                expr = " ".join(parsed_inline.remaining)
            _execute_expression(
                expr,
                format_mode=repl_format_mode,
                relaxed=repl_relaxed,
                simplify_output=repl_simplify_output,
                explain_parse=repl_explain_parse,
                always_wa=repl_always_wa,
                copy_wa=repl_copy_wa,
                color_mode=repl_color_mode,
                session_locals=session_locals,
            )
        except (EOFError, KeyboardInterrupt):
            print()
            return 0
        except Exception as exc:
            _print_error(exc, expr=expr, color_mode=repl_color_mode, session_locals=session_locals)


if __name__ == "__main__":
    raise SystemExit(run())



from __future__ import annotations

from sympy import Eq, Symbol, dsolve
from sympy.core.function import AppliedUndef
from sympy.logic.boolalg import Boolean


def split_top_level_commas(text: str) -> list[str]:
    parts: list[str] = []
    current: list[str] = []
    depth = 0
    for ch in text:
        if ch == "(":
            depth += 1
        elif ch == ")":
            depth = max(0, depth - 1)
        if ch == "," and depth == 0:
            piece = "".join(current).strip()
            if piece:
                parts.append(piece)
            current = []
            continue
        current.append(ch)
    tail = "".join(current).strip()
    if tail:
        parts.append(tail)
    return parts


def infer_ode_dependent(eq_value: Eq):
    candidates = sorted(eq_value.atoms(AppliedUndef), key=str)
    if not candidates:
        return None
    return candidates[0]


def _reject_mixed_dependent_notation(value, dep) -> None:
    dep_name = getattr(dep.func, "__name__", None)
    if not dep_name:
        return
    if Symbol(dep_name) in value.free_symbols:
        raise ValueError(
            f"mixed dependent variable notation: found both {dep_name} and {dep_name}(x); use one style"
        )


def evaluate_ode_alias(
    expr: str,
    *,
    evaluate_fn,
    relaxed: bool,
    simplify_output: bool,
    session_locals: dict | None = None,
):
    body = expr[4:].strip()
    if not body:
        raise ValueError("ode expects an equation, e.g. ode y' = y")

    pieces = split_top_level_commas(body)
    if not pieces:
        raise ValueError("ode expects an equation, e.g. ode y' = y")
    equation_text, ic_texts = pieces[0], pieces[1:]

    eq_value = evaluate_fn(
        equation_text,
        relaxed=relaxed,
        session_locals=session_locals,
        simplify_output=simplify_output,
    )
    if not isinstance(eq_value, Eq):
        raise ValueError("ode expects an equation, e.g. ode y' = y")

    dep = infer_ode_dependent(eq_value)
    if dep is None:
        raise ValueError("could not infer dependent function; use y(x)-style notation")
    _reject_mixed_dependent_notation(eq_value, dep)

    ics: dict = {}
    for ic_text in ic_texts:
        ic_value = evaluate_fn(
            ic_text,
            relaxed=relaxed,
            session_locals=session_locals,
            simplify_output=simplify_output,
        )
        if not isinstance(ic_value, Eq):
            if isinstance(ic_value, (bool, Boolean)):
                raise ValueError(
                    f"initial condition reduced to a boolean: {ic_text}; use forms like y(0)=1 or y'(0)=0"
                )
            raise ValueError(f"initial condition must be an equation: {ic_text}")
        _reject_mixed_dependent_notation(ic_value, dep)
        ics[ic_value.lhs] = ic_value.rhs

    if ics:
        result = dsolve(eq_value, dep, ics=ics)
        ics_rendered = ", ".join(f"{lhs}: {rhs}" for lhs, rhs in ics.items())
        parsed_expr = f"dsolve({eq_value}, {dep}, ics={{{ics_rendered}}})"
    else:
        result = dsolve(eq_value, dep)
        parsed_expr = f"dsolve({eq_value}, {dep})"
    return result, parsed_expr



import calc.updates as updates


class _DummyResponse:
    def __init__(self, payload: bytes):
        self._payload = payload

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        return False

    def read(self) -> bytes:
        return self._payload


def test_latest_pypi_version_success():
    def fake_urlopen(url: str, timeout: float):
        assert "pypi.org/pypi/philcalc/json" in url
        assert timeout == 2.0
        return _DummyResponse(b'{"info":{"version":"1.2.3"}}')

    assert updates.latest_pypi_version("philcalc", urlopen_fn=fake_urlopen) == "1.2.3"


def test_latest_pypi_version_failures():
    def os_error(*args, **kwargs):
        raise OSError("offline")

    assert updates.latest_pypi_version("philcalc", urlopen_fn=os_error) is None

    def bad_json(*args, **kwargs):
        return _DummyResponse(b"{")

    assert updates.latest_pypi_version("philcalc", urlopen_fn=bad_json) is None


def test_compare_versions_all_branches():
    assert updates.compare_versions("1.2.3", "1.2.3") == 0
    assert updates.compare_versions("1.2.3", "1.2.4") == -1
    assert updates.compare_versions("1.2.4", "1.2.3") == 1
    assert updates.compare_versions("1.2.3", "1.2.3.dev1") == 1
    assert updates.compare_versions("1.2.3.dev1", "1.2.3") == -1
    assert updates.compare_versions("1.2.3.dev1", "1.2.3.dev2") == -1
    assert updates.compare_versions("1.2.3.dev3", "1.2.3.dev2") == 1
    assert updates.compare_versions("1.2.3.dev2", "1.2.3.dev2") == 0
    assert updates.compare_versions("wat", "1.2.3") is None


def test_update_status_lines_branches():
    cmd = "uv tool upgrade philcalc"
    assert updates.update_status_lines("dev", None, cmd)[0] == "current version: dev (local checkout)"

    unavailable = updates.update_status_lines("1.2.3", None, cmd)
    assert "latest version: unavailable" in unavailable[1]

    up_to_date = updates.update_status_lines("1.2.3", "1.2.3", cmd)
    assert "up to date" in up_to_date[1]
    assert "no update needed" in up_to_date[2]

    update_available = updates.update_status_lines("1.2.3", "1.2.4", cmd)
    assert "update available" in update_available[1]
    assert update_available[2] == f"update with: {cmd}"

    newer_local = updates.update_status_lines("1.2.3.dev1", "1.2.2", cmd)
    assert "newer local/pre-release build" in newer_local[1]

    unknown_compare = updates.update_status_lines(
        "1.2.3",
        "1.2.4",
        cmd,
        compare_fn=lambda a, b: None,
    )
    assert "version comparison unavailable" in unknown_compare[1]
    assert unknown_compare[2] == f"update with: {cmd}"


def test_repl_startup_update_status_lines_branches():
    cmd = "uv tool upgrade philcalc"
    assert updates.repl_startup_update_status_lines("dev", "1.2.3", cmd) == [
        "startup update check: dev (local checkout)"
    ]
    assert updates.repl_startup_update_status_lines("1.2.3", None, cmd) == [
        "startup update check: latest version unavailable"
    ]
    assert updates.repl_startup_update_status_lines("1.2.3", "1.2.3", cmd) == [
        "startup update check: v1.2.3 is up to date"
    ]
    assert updates.repl_startup_update_status_lines("1.2.3", "1.2.4", cmd) == [
        "startup update check: v1.2.4 available (you have v1.2.3)",
        f"update with: {cmd}",
    ]
    assert updates.repl_startup_update_status_lines("1.2.4", "1.2.3", cmd) == [
        "startup update check: v1.2.4 is newer than latest release v1.2.3"
    ]
    assert updates.repl_startup_update_status_lines(
        "1.2.3",
        "1.2.4",
        cmd,
        compare_fn=lambda a, b: None,
    ) == ["startup update check: version comparison unavailable"]



from __future__ import annotations

from dataclasses import dataclass

FORMAT_MODES = ("plain", "pretty", "latex", "latex-inline", "latex-block", "json")
COLOR_MODES = {"auto", "always", "never"}


@dataclass(frozen=True)
class CLIOptions:
    format_mode: str = "plain"
    relaxed: bool = True
    simplify_output: bool = True
    explain_parse: bool = False
    always_wa: bool = False
    copy_wa: bool = False
    color_mode: str = "auto"
    remaining: tuple[str, ...] = ()


def parse_options(args: list[str], *, help_text: str) -> CLIOptions:
    format_mode = "plain"
    relaxed = True
    simplify_output = True
    explain_parse = False
    always_wa = False
    copy_wa = False
    color_mode = "auto"
    idx = 0
    while idx < len(args) and args[idx].startswith("-"):
        arg = args[idx]
        if arg in {"-h", "--help"}:
            print(help_text)
            raise SystemExit(0)
        if arg == "--format":
            if idx + 1 >= len(args):
                raise ValueError("missing value for --format")
            mode = args[idx + 1]
            if mode not in FORMAT_MODES:
                raise ValueError(f"unknown format mode: {mode}")
            format_mode = mode
            idx += 2
            continue
        if arg.startswith("--format="):
            mode = arg.split("=", 1)[1]
            if mode not in FORMAT_MODES:
                raise ValueError(f"unknown format mode: {mode}")
            format_mode = mode
            idx += 1
            continue
        if arg == "--latex":
            format_mode = "latex"
            idx += 1
            continue
        if arg == "--latex-inline":
            format_mode = "latex-inline"
            idx += 1
            continue
        if arg == "--latex-block":
            format_mode = "latex-block"
            idx += 1
            continue
        if arg == "--strict":
            relaxed = False
            idx += 1
            continue
        if arg == "--no-simplify":
            simplify_output = False
            idx += 1
            continue
        if arg == "--explain-parse":
            explain_parse = True
            idx += 1
            continue
        if arg == "--wa":
            always_wa = True
            idx += 1
            continue
        if arg == "--copy-wa":
            copy_wa = True
            idx += 1
            continue
        if arg == "--color":
            if idx + 1 >= len(args):
                raise ValueError("missing value for --color")
            mode = args[idx + 1]
            if mode not in COLOR_MODES:
                raise ValueError(f"unknown color mode: {mode}")
            color_mode = mode
            idx += 2
            continue
        if arg.startswith("--color="):
            mode = arg.split("=", 1)[1]
            if mode not in COLOR_MODES:
                raise ValueError(f"unknown color mode: {mode}")
            color_mode = mode
            idx += 1
            continue
        if arg == "--":
            idx += 1
            break
        if arg.startswith("--"):
            raise ValueError(f"unknown option: {arg}")
        break

    return CLIOptions(
        format_mode=format_mode,
        relaxed=relaxed,
        simplify_output=simplify_output,
        explain_parse=explain_parse,
        always_wa=always_wa,
        copy_wa=copy_wa,
        color_mode=color_mode,
        remaining=tuple(args[idx:]),
    )



from __future__ import annotations

import json
import re
from urllib.request import urlopen

_SEMVERISH_PATTERN = re.compile(r"^(\d+)\.(\d+)\.(\d+)(?:\.dev(\d+))?$")


def latest_pypi_version(package_name: str, *, urlopen_fn=urlopen) -> str | None:
    url = f"https://pypi.org/pypi/{package_name}/json"
    try:
        with urlopen_fn(url, timeout=2.0) as response:
            payload = json.loads(response.read().decode("utf-8"))
        return payload.get("info", {}).get("version")
    except (OSError, TimeoutError, ValueError):
        return None


def compare_versions(current: str, latest: str) -> int | None:
    current_match = _SEMVERISH_PATTERN.match(current)
    latest_match = _SEMVERISH_PATTERN.match(latest)
    if current_match is None or latest_match is None:
        return None

    current_release = tuple(int(part) for part in current_match.group(1, 2, 3))
    latest_release = tuple(int(part) for part in latest_match.group(1, 2, 3))
    if current_release < latest_release:
        return -1
    if current_release > latest_release:
        return 1

    current_dev = current_match.group(4)
    latest_dev = latest_match.group(4)
    if current_dev is None and latest_dev is None:
        return 0
    if current_dev is None:
        return 1
    if latest_dev is None:
        return -1

    current_dev_num = int(current_dev)
    latest_dev_num = int(latest_dev)
    if current_dev_num < latest_dev_num:
        return -1
    if current_dev_num > latest_dev_num:
        return 1
    return 0


def update_status_lines(
    version: str,
    latest: str | None,
    update_cmd: str,
    *,
    compare_fn=compare_versions,
) -> list[str]:
    if version == "dev":
        return [
            "current version: dev (local checkout)",
            "latest version: unknown from local checkout",
            "install latest local changes with: uv tool install --force --reinstall --refresh .",
        ]

    lines = [f"current version: {version}"]
    if latest is None:
        lines.append("latest version: unavailable (offline or PyPI unreachable)")
        lines.append("hint: retry :check when online")
        return lines

    relation = compare_fn(version, latest)
    if relation == 0 or latest == version:
        lines.append(f"latest version: {latest} (up to date)")
        lines.append("no update needed")
    elif relation == -1:
        lines.append(f"latest version: {latest} (update available)")
        lines.append(f"update with: {update_cmd}")
    elif relation == 1:
        lines.append(f"latest version: {latest} (you are on a newer local/pre-release build)")
        lines.append("no update needed")
    else:
        lines.append(f"latest version: {latest} (version comparison unavailable)")
        lines.append(f"update with: {update_cmd}")
    return lines


def repl_startup_update_status_lines(
    version: str,
    latest: str | None,
    update_cmd: str,
    *,
    compare_fn=compare_versions,
) -> list[str]:
    if version == "dev":
        return ["startup update check: dev (local checkout)"]
    if latest is None:
        return ["startup update check: latest version unavailable"]

    relation = compare_fn(version, latest)
    if relation == 0 or latest == version:
        return [f"startup update check: v{version} is up to date"]
    if relation == -1:
        return [
            f"startup update check: v{latest} available (you have v{version})",
            f"update with: {update_cmd}",
        ]
    if relation == 1:
        return [f"startup update check: v{version} is newer than latest release v{latest}"]
    return ["startup update check: version comparison unavailable"]



from __future__ import annotations

import shlex
import sys
from typing import Callable

from .options import CLIOptions


def handle_repl_command(
    expr: str,
    *,
    help_text: str,
    help_chain_text: str,
    help_power_text: str,
    help_demo_text: str,
    examples_text: str,
    tutorial_text: str,
    ode_text: str,
    linalg_text: str,
    cli_name: str,
    version: str,
    print_update_status: Callable[[], None],
    style_fn: Callable[..., str],
    color_mode: str = "auto",
    stderr=sys.stderr,
) -> bool:
    if expr in {":q", ":quit", ":x"}:
        raise EOFError
    if expr in {":h", ":help"}:
        print(help_text)
        return True
    if expr == "?":
        print(help_text)
        print(help_chain_text)
        return True
    if expr == "??":
        print(help_power_text)
        return True
    if expr == "???":
        print(help_demo_text)
        return True
    if expr == ":examples":
        print(examples_text)
        return True
    if expr in {":tutorial", ":tour"}:
        print(tutorial_text)
        return True
    if expr == ":ode":
        print(ode_text)
        return True
    if expr in {":linalg", ":la"}:
        print(linalg_text)
        return True
    if expr in {":v", ":version"}:
        print(f"{cli_name} v{version}")
        return True
    if expr in {":update", ":check"}:
        print_update_status()
        return True
    if expr.startswith(":"):
        print(style_fn("E: unknown command", color="red", stream=stderr, color_mode=color_mode), file=stderr)
        print(
            style_fn("hint: use :h to list commands", color="yellow", stream=stderr, color_mode=color_mode),
            file=stderr,
        )
        return True
    return False


def tutorial_command(
    expr: str,
    state: dict | None,
    *,
    tutorial_steps: tuple[str, ...],
    print_tutorial_step: Callable[[int], None],
    stderr=sys.stderr,
) -> bool:
    if state is None:
        return False
    if expr in {":tutorial", ":tour"}:
        state["active"] = True
        state["index"] = 0
        print("tutorial mode started. use :next, :repeat, :done")
        print_tutorial_step(state["index"])
        return True
    if expr == ":next":
        if not state.get("active", False):
            print("hint: start with :tutorial", file=stderr)
            return True
        nxt = state["index"] + 1
        if nxt >= len(tutorial_steps):
            print("tutorial complete. use :done to exit tutorial mode")
            return True
        state["index"] = nxt
        print_tutorial_step(state["index"])
        return True
    if expr == ":repeat":
        if not state.get("active", False):
            print("hint: start with :tutorial", file=stderr)
            return True
        print_tutorial_step(state["index"])
        return True
    if expr == ":done":
        if state.get("active", False):
            state["active"] = False
            print("tutorial mode ended")
        else:
            print("hint: tutorial is not active; use :tutorial", file=stderr)
        return True
    return False


def try_parse_repl_inline_options(
    expr: str,
    *,
    cli_name: str,
    parse_options_fn: Callable[[list[str]], CLIOptions],
) -> CLIOptions | None:
    line = expr.strip()
    if line.startswith(f"{cli_name} "):
        line = line[len(cli_name) :].strip()
    elif not line.startswith("-"):
        return None
    try:
        tokens = shlex.split(line)
    except ValueError as exc:
        raise ValueError(f"invalid REPL option input: {exc}") from exc
    return parse_options_fn(tokens)
